/*
==dbapi20==
  Prototype for dbapi20 modules.

====Example====
 var db=new mysql("localhost","user","password","database");
 // or
 var db=open("mysql://user:password@localhost/database");
 // same result

 var people=db.query("SELECT * FROM people");
 var i;
 for (i in people) {
 	print(people[i].id, people[i].name,"\n");
 }

 var name=db.query("SELECT name FROM people");
 print(name[0],"\n");

 db.query("INSERT INTO people (id, name) VALUES (:id, :name)", {id: 42, name:"John"});
*/


dbapi20=function() {
}

/*

====query(qry, [parameters])====

Creates a cursor, passes the query to the server and
fetches all rows in the result using cursor.fetchall().
Closes the cursor.

*/

dbapi20.prototype.query=function(qry, parameters) {
	var cur=this.cursor();
	cur.execute.apply(cur,arguments);
	var ret=cur.fetchall();
	cur.close();
	return ret;
}

/*
Parent class for cursors.
Provides some default functions like fetchone(), fetchmany() and fetchall(), so you don't
have to implement all three
*/

dbapi20.cursor=function() {}

/*
        ====executemany(operation,array_of_parameters)====
          
            Implemented using multiple calls to execute.

            The same comments as for execute() also apply accordingly
            to this method.
            
            Return values are not defined.
*/

dbapi20.cursor.prototype.executemany=function(operation,array_of_parameters) {
	var i;
	for (i in array_of_parameters) {
		this.execute(operation,array_of_parameters[i]);
	}
}

/*
        ====execute(operation[,parameters ...])====
          
            Prepare and execute a database operation (query or
            command).  Parameters may be provided as sequence or
            mapping and will be bound to variables in the operation.

	operation may contain parameters, like "select * from users where userid=:paramname".
	parameters should not have names when parameters are provided as additional arguments.

            Return values are not defined.
*/
dbapi20.cursor.prototype.execute=function(operation,parameters) {
	this.executemany(operation,[parameters]);
}

/*
        ====fetchone() ====
          
            Fetch the next row of a query result set, returning a
            single value or object, or null when no more data is
            available.

	If the query results in zero or more than one columns, the return value is
	an object which contains properties with the same name as the column names.
	Values are converted into strings, numbers or dates.

	If the query results in a single column, that column is returned as a value
	instead of a single-property object.

*/

dbapi20.cursor.prototype.fetchone=function() {
	var ret=this.fetchmany(1);
	if (!ret) return null;
	return ret[0];
}

/*
        ====fetchmany([size=cursor.arraysize])====
          
            Fetch the next set of rows of a query result, returning an
            array of objects. null
            is returned when no more rows are available.
            
            The number of rows to fetch per call is specified by the
            parameter.  If it is not given, the cursor's arraysize
            determines the number of rows to be fetched. The method
            tries to fetch as many rows as indicated by the size
            parameter. If this is not possible due to the specified
            number of rows not being available, fewer rows may be
            returned.

	If size=-1, will fetch all remaining rows
            
*/

dbapi20.cursor.prototype.fetchmany=function(size) {
	if (arguments.length<1) size=this.arraysize;
	var ret=[];
	var i;
	for (i=0; size==-1 || i<size; i++) {
		var one=this.fetchone();
		if (one==null) break;
		ret.push(one);
	}
	if (ret.length==0) return null;
	return ret;
}

/*           
        ====fetchall() ====

            Fetch all (remaining) rows of a query result, returning
            them as an array of objects.
*/

dbapi20.cursor.prototype.fetchall=function() {
	var ret=this.fetchmany(-1);
	return ret;
}

/*
        ====arraysize====
          
            This read/write attribute specifies the number of rows to
            fetch at a time with fetchmany(). It defaults to 1 meaning
            to fetch a single row at a time.
            
            Implementations must observe this value with respect to
            the fetchmany() method, but are free to interact with the
            database a single row at a time. It may also be used in
            the implementation of executemany().

*/

dbapi20.cursor.prototype.arraysize=1

/*
    ====rownumber====

        This read-only attribute is the current 0-based
        index of the cursor in the result set or null if the index cannot
        be determined.

        The index can be seen as index of the cursor in a sequence (the
        result set). The next fetch operation will fetch the row
        indexed by .rownumber in that sequence.

*/

dbapi20.cursor.prototype.rownumber=null;

