if (0!=_.FCGX_IsCGI()) throw("fcgi: isCgi");
if (0!=_.FCGX_Init()) throw("fcgi: Init");

function fcgi(path, backlog) {
	if (backlog==undefined) backlog=8;
	if (path==undefined) {
		this.fd=0;
	} else {
		this.fd=_.FCGX_OpenSocket(path,backlog);
		if (this.fd==-1) throw("fcgi:OpenSocket");
	}

	this.request=_.New(_.struct.FCGX_Request);
	var res=_.FCGX_InitRequest(this.request, this.fd, 0);
	if (res!=0) throw("fcgi:InitRequest");
}

fcgi.server=function(path,backlog) {
	var request=new fcgi(path,backlog);
	while(acc=request.accept()) {
		if (environment.JSEXT_FCGI_FORK!=undefined) {
			var pid=os.fork();
		} else {
			var pid=0;
		}
		if (!pid) { // Child
			stdin=acc[0];
			stdout=acc[1];
			stderr=acc[2];
			environment=acc[3];

			cgi.init();

			if (global.arguments.length>0) { // Dynamic fcgi app
				var scriptfile=global.arguments[0];
			} else { // Static fcgi app
				var scriptfile=environment.DOCUMENT_ROOT+environment.SCRIPT_NAME;
				var lastslash=scriptfile.lastIndexOf("/");
				if (lastslash!=-1) {
					os.chdir(scriptfile.substr(0,lastslash));
					scriptfile=scriptfile.substr(lastslash+1);
				}
			}
		} else {
			os.waitpid(pid);
		}

		try {
			load(scriptfile);
		} catch (err) {
			print("Line ",err.lineNumber," in ",err.fileName," while loading ",scriptfile,": ",err,"\n");
			print("cwd=",os.getcwd());
		}
		request.finish();
	}
}

fcgi.prototype.accept=function() {
	var res=_.FCGX_Accept_r(this.request);
	if (res==0) return [
		new fcgi.file(this.request["in"]),
		new fcgi.file(this.request.out),
		new fcgi.file(this.request.err),
		this.environment()
	];
}

fcgi.prototype.finish=function() {
	_.FCGX_Finish_r(this.request);
}

fcgi.prototype.free=function(close) {
	if (close==undefined) close=1;
	var res=_.FCGX_Free(this.request,close);
	if (res!=0) throw("fcgi:Free");
	delete this.request;
}

fcgi.prototype.environment=function() {
	var ret={};
	var envp=this.request.envp;
	if (envp==null) return;
	var i;
	for(i=0;;i++) {
		var val=envp[i];
		if (val==null) break;
		val=val.string();
		var eqPos=val.indexOf("=");
		ret[val.substr(0,eqPos)]=val.substr(eqPos+1);
	}
	return ret;
}

fcgi.file=function(stream) {
	this.stream=stream;
	this.closed=false;
}

fcgi.file.prototype.closed=true;

fcgi.file.prototype.close=function() {
	if (!this.closed) {
		_.FCGX_FClose(this.stream);
		this.closed=true;
	}
}

fcgi.file.prototype.eof=function() {
	return _.FCGX_HasSeenEOF(this.stream)?true:false;
}

fcgi.file.prototype.flush=function() {
	_.FCGX_FFlush(this.stream);
}

fcgi.file.prototype.isatty=function() {
	return false;
}

fcgi.file.prototype.read=function(size) {
	if (arguments.length<1) size=-1;
	if (size<0) {
		var trysize=512;
		var ret="";
		while (!this.eof()) {
			ret+=this.read(trysize);
			trysize*=2;
		}
		return ret;
	}
	var buf=new Pointer(size);
	var len=_.FCGX_GetStr(buf, size, this.stream);
	return buf.string(len);
}

// Not binary-safe

fcgi.file.prototype.readline=function(size) {
	if (arguments.length<1) size=-1;
	if (size<0) {
		var trysize=512;
		var ret="";
		while (!this.eof()) {
			ret+=this.readline(trysize);
			if (ret[ret.length-1]=='\n') break;
			trysize*=2;
		}
		return ret;
	}
	var buf=new Pointer(size+1);
	var beforepos=this.tell();
	var line=_.FCGX_GetLine(buf, size+1, this.stream);
	if (line==null) return "";
	return buf.string();
}

fcgi.file.prototype.readlines=function(size) {
	if (arguments.length<1) size=-1;
	if (size<0) {
		var ret=[];
		while (!this.eof()) {
			ret.push(this.readline());
		}
		return ret;
	}
	var buf=this.read(size);
	return buf.split(/\n/);
}

fcgi.file.prototype.write=function(str) {
	_.FCGX_PutStr(str, str.length, this.stream);
}

fcgi.file.prototype.writelines=function(obj) {
	for (i in obj) {
		this.write(obj[i]);
	}
}

