function file(fp, bufsize) {
	if (arguments.length>0) {
		if (fp==null) throw('file: Could not open file');

		this.fp=fp;
		this.closed=false;
	}

	if (bufsize!=undefined) {
		_.setvbuf(fp,null,_._IOFBF,bufsize);
	}
}

/*
====open(filename, [mode, [bufsize]])====

Returns an open [[file]] object or throws an error.
*/

file.open=function(filename, mode, bufsize) {
	if (filename[0]=='|') {
		if (mode==undefined) mode="w";
		return os.popen(filename.substr(1),mode,bufsize);
	} else if (filename[filename.length-1]=='|') {
		return os.popen(filename.substr(0,filename.length-1),mode,bufsize);
	}

	if (mode==undefined) mode="r";

	var ret=new file();
	ret.fp=_.fopen(filename,mode);
	if (ret.fp==null) os.error('file.open: Failed to open file \''+filename+'\'');
	ret.name=filename;
	ret.closed=false;

	if (bufsize) _.setvbuf(ret.fp,null,_._IOFBF,bufsize);

	return ret;
}

/**/

// used by open.js

file.open_uri=function(uri, mode, bufsize) {
	return this.open(uri.path, mode, bufsize);
}

file.prototype.closed=true;

/*

*/

file.prototype.close=function() {
	if (!this.closed) {
		_.fclose(this.fp);
		this.closed=true;
	}
}

/*

*/

// eof is not part of python
file.prototype.eof=function() {
	return _.feof(this.fp)?true:false;
}

/*

*/

file.prototype.flush=function() {
	_.fflush(this.fp);
}

/*

*/

file.prototype.fileno=function() {
	return _.fileno(this.fp);
}

/*

*/

file.prototype.isatty=function() {
	return _.isatty(this.fileno());
}

/*

*/

file.prototype.read=function(size) {
	if (arguments.length<1) size=-1;
	if (size<0) {
		var trysize=512;
		var ret="";
		while (!this.eof()) {
			ret+=this.read(trysize);
			trysize*=2;
		}
		return ret;
	}
	var buf=new Pointer(size);
	var len=_.fread(buf, 1, size, this.fp);
	return buf.string(len);
}

/*

Not Binary-safe if operating system's fgets reads null characters

*/

file.prototype.readline=function(size) {
	if (arguments.length<1) size=-1;
	if (size<0) {
		var trysize=512;
		var ret="";
		while (!this.eof()) {
			ret+=this.readline(trysize);
			if (ret[ret.length-1]=='\n') break;
			trysize*=2;
		}
		return ret;
	}
	var buf=new Pointer(size+1);
//	var beforepos=this.tell();
	var line=_.fgets(buf, size+1, this.fp);
	return buf.string();//(this.tell()-beforepos);
}

/*

*/

file.prototype.readlines=function(size) {
	if (arguments.length<1) size=-1;
	if (size<0) {
		var ret=[];
		while (!this.eof()) {
			ret.push(this.readline());
		}
		return ret;
	}
	var buf=this.read(size);
	return buf.split(/\n/);
}

file.SEEK_SET=_.SEEK_SET;
file.SEEK_CUR=_.SEEK_CUR;
file.SEEK_END=_.SEEK_END;

/*
===seek(offset, whence=file.SEEK_SET)===

May use file.SEEK_SET, file.SEEK_CUR or file.SEEK_END.
*/

file.prototype.seek=function(offset, whence) {
	if (arguments.length<2) whence=file.SEEK_SET;
	_.clearerr(this.fp);
	_.fseek(this.fp, offset, whence);
}

/*

*/

file.prototype.tell=function() {
	return _.ftell(this.fp);
}

/*

*/

file.prototype.truncate=function(size) {
	_.ftruncate(this.fileno(),size);
}

/*

*/

file.prototype.write=function(str) {
	return _.fwrite(str, 1, str.length, this.fp);
}

/*

*/

file.prototype.writelines=function(obj) {
	for (i in obj) {
		this.write(obj[i]);
	}
}


