function gdbm(file, mode, block_size, options) {
	if (mode==undefined) mode="r";
	if (block_size==undefined) block_size=0;
	if (options==undefined) options=0;

	this.file=file;
	switch(mode) {
	case "r": var read_write=_.GDBM_READER; break;
	case "w": var read_write=_.GDBM_WRCREAT; break;
	default: throw("gdbm: Wrong mode '"+mode+"'");
	}
	this.dbf=_.gdbm_open(file, block_size, read_write | options, 0777, null);
	if (this.dbf==null) throw("gdbm: open '"+file+"' failed");
}

gdbm.SYNC=_.GDBM_SYNC;
gdbm.NOLOCK=_.GDBM_NOLOCK;

gdbm.open_uri=function(uri, mode, block_size, options) {
	return new gdbm(uri.path, mode, block_size, options);
}

gdbm.prototype.close=function() {
	_.gdbm_close(this.dbf);
}

gdbm.INSERT=_.GDBM_INSERT;
gdbm.REPLACE=_.GDBM_REPLACE;

gdbm.prototype.store=function(key, content, flag) {
	if (flag==undefined) flag=_.GDBM_REPLACE;
	var res=_.gdbmw_store(this.dbf, key, key.length, content, content.length, flag);
	if (res==-1) throw("gdbm: insert");
	return res?true:false;
}

gdbm.prototype.fetch=function(key) {
	var rlen=[0];
	var retp=_.gdbmw_fetch(this.dbf, key, key.length, rlen);
	var ret=retp.string(rlen[0]);
	_.free(retp);
	return ret;
}

gdbm.prototype._delete=function(key) {
	var res=_.gdbmw_delete(this.dbf, key, key.length);
	return res?false:true; // 0 is success
}

gdbm.prototype.firstkey=function() {
	var rlen=[0];
	var retp=_.gdbmw_firstkey(this.dbf, rlen);
	if (retp==null) return; // null value for no key
	var ret=retp.string(rlen[0]);
	_.free(retp);
	return ret;
}

gdbm.prototype.nextkey=function(key) {
	var rlen=[0];
	var retp=_.gdbmw_nextkey(this.dbf, key, key.length, rlen);
	if (retp==null) return; // null value for no key
	var ret=retp.string(rlen[0]);
	_.free(retp);
	return ret;
}

gdbm.prototype.reorganize=function() {
	return _.gdbm_reorganize(this.dbf);
}

gdbm.prototype.sync=function() {
	_.gdbm_sync(this.dbf);
}

/*
====exists(key)====
*/

gdbm.prototype.exists=function(key) {
	var res=_.gdbmw_exists(this.dbf, key, key.length);
	return res?true:false;
}

/*
====set(key, value, [replace=true])====
*/

gdbm.prototype.set=function(key, value, replace) {
	if (replace==undefined) replace=true;
	if (replace==true) {
		return this.store(key,value,this.REPLACE);
	} else {
		return this.store(key,value,this.INSERT);
	}
}

/*
====get(key)====
*/

gdbm.prototype.get=gdbm.prototype.fetch;

/*
====read()====

Reads all values in database and returns an object with the same
key/value associations. All values are strings.
*/

gdbm.prototype.read=function() {
	var ret={};

	this.reset();
	var pair;
	while ((pair=this.each())) {
		ret[pair[0]]=pair[1];
	}
	return ret;
}


/*
====write(obj)====

Writes properties of obj to database. All properties should be strings.
*/

gdbm.prototype.write=function(obj) {
	var i;
	for (i in obj) {
		this.set(i,obj[i]);
	}
}

/*
====reset()====

Reset iterator. Use with [[cdb.each]].
*/

gdbm.prototype.reset=function() {
	this.current_key=this.firstkey();
}

/*
====each()====

Return [key, value] or ''undefined'' and advance iterator
*/

gdbm.prototype.each=function() {
	if (this.current_key==undefined) return;
	var ret=[this.current_key,this.get(this.current_key)];
	this.current_key=this.nextkey(this.current_key);
	return ret;
}


