function webpage(str) {
	if (str!=undefined)
		this.page=str;
	else
		this.page="";
}

webpage.prototype.toString=function () {
	return this.page;
}

/*
====get_inside(id)====

Returns a subpage containing only the innards if the tag with 'id''.
*/

webpage.prototype.get_inside=function(id) {
	var pos=this.get_inside_pos(id);
	return new webpage(this.page.substr(pos[0],pos[1]));
}


/*
====get_outside(id)====

Returns a subpage containing only the the tag with ''id'' and its contents.
*/

webpage.prototype.get_inside=function(id) {
	var pos=this.get_inside_pos(id);
	return new webpage(this.page.substr(pos[0],pos[1]));
}


/*
*/

webpage.prototype.get_pos_by_id=function(id) {
	var newpos=this.page.indexOf('id="'+id+'"');
	if (newpos==-1) throw('webpage: Could not find id="'+id+'"');
	var tagOpen=this.page.lastIndexOf("<",newpos);
	if (tagOpen==-1) throw("webpage: Malformed HTML (1)");
	return tagOpen;
}


/*
====get_outside_pos(Number pos)====

Returns position and length of tag which opens at ''pos'' along with its contents and end tag.

====get_outside_pos(String id)====

Returns position and length of tag with the given id.

Requires tag to have an end tag and does not
accept xml-style empty tags: <pre><br/></pre>.
*/

webpage.prototype.get_outside_pos=function(pos) {
	if (typeof(pos)=="string") {
		pos=this.get_pos_by_id(pos);
	}
	var tagClose=this.page.indexOf(">",pos);
	if (tagClose==-1) throw("webpage: Malformed HTML (2)");
	var openTag=this.page.substr(pos,tagClose-pos);
	var tag=openTag.match(/<([a-zA-Z0-9]*)/);
	if (!tag) throw("webpage: Malformed HTML (3)");
	tag=tag[1];

	var startTag="<"+tag;
	var endTag="</"+tag+">";

	var level=1; // at beginpos
	var atPos=pos;

// 1 2 3 2 3 2 1 0
// ( ( ( ) ( ) ) )

	var beginPos=this.page.indexOf(startTag,atPos+1);
	var endPos=this.page.indexOf(endTag,atPos+1);

	for(;;) {
		if (endPos==-1) throw("webpage: Malformed HTML (4)");
		if (beginPos<endPos && beginPos!=-1) {
			level++;
			atPos=beginPos;			
			beginPos=this.page.indexOf(startTag,atPos+1);
		} else {
			level--;
			if (level==0) break;
			atPos=endPos;
			endPos=this.page.indexOf(endTag,atPos+1);
		}
	}
	return [pos,endPos+endTag.length-pos];
}

/*
====get_inside_pos(Number pos)====

Returns position and length of tag which opens at ''pos''

====get_inside_pos(String id)====

Returns position and length of the tag with the given id.

Requires tag to have an end tag and does not
accept xml-style empty tags: <pre><br/></pre>.
*/

webpage.prototype.get_inside_pos=function(pos) {
	var ret=this.get_outside_pos(pos);
	var endPos=this.page.lastIndexOf("<",ret[0]+ret[1]-1);
	var startpos=this.page.indexOf('>',ret[0])+1;
	return [startpos,endPos-startpos];
}

/*
====replace_outside(id,value)====

Replaces first occurence of a tag with id=''id'' with ''value''. The whole tag,
including its contents and end tag is replaced.
*/

webpage.prototype.replace_outside=function(id,value) {
	var atpos=0;
	var ret=[];

	var wholeTag=this.get_outside_pos(id);
	ret.push(this.page.substr(atpos,wholeTag[0]));
	ret.push(value);
	ret.push(this.page.substr(atpos+wholeTag[0]+wholeTag[1]));
	this.page=ret.join("");
}

/*
====replace_inside(id,value)====

Replaces inside of the first occurence of a tag with id=''id'' with ''value''. Keeps the
actual tag and closing tag, replaces its inside.
*/

webpage.prototype.replace_inside=function(id,value) {
	var atpos=0;
	var ret=[];

	var wholeTag=this.get_inside_pos(id);
	ret.push(this.page.substr(atpos,wholeTag[0]));
	ret.push(value);
	ret.push(this.page.substr(atpos+wholeTag[0]+wholeTag[1]));
	this.page=ret.join("");
}


/**/

webpage.prototype.get_attribute_pos=function(id, attribute) {
	var pos=this.get_pos_by_id(id);
	var attribstr=attribute+'="';
	var attribpos=this.page.indexOf(attribstr,pos);
	var attribend=this.page.indexOf('"',attribpos+attribstr.length);
	if (attribpos==-1 || attribend==-1) throw("webpage: Could not find attribute '"+attribute+"'");

	attribpos+=attribstr.length;

	return [attribpos, attribend-attribpos];
}

/*
====replace_attribute(id,attribute,value)====

Replaces value of first occurence of an attribute named ''attribute'' in a tag with id=''id'' with ''value''.
Attribute must occur like ''attribute="[optional value]"''.
*/

webpage.prototype.replace_attribute=function(id,attribute,value) {
	var attribPos=this.get_attribute_pos(id,attribute);

	this.page=[
		this.page.substr(0,attribPos[0]),
		value,
		this.page.substr(attribPos[0]+attribPos[1])
		].join("");
}

/*
====get_attribute(id,attribute)====

*/

webpage.prototype.get_attribute=function(id,attribute) {
	var attribPos=this.get_attribute_pos(id,attribute);

	return this.page.substr(attribPos[0],attribPos[1]);
}


/*
====replace_editable_outside(id,value)====

Replaces first occurence of an editable area with id=''id'' with ''value''. The whole area,
including its surrounding meta-comments is replaced.
*/

webpage.prototype.replace_editable_outside=function(id,value) {
	var atpos=0;
	var ret=[];

	var wholeTag=this.get_editable_outside_pos(id);
	ret.push(this.page.substr(atpos,wholeTag[0]));
	ret.push(value);
	ret.push(this.page.substr(atpos+wholeTag[0]+wholeTag[1]));
	this.page=ret.join("");
}

/*
====replace_editable_inside(id,value)====

Replaces the first occurence of an editable area with id=''id'' with ''value''. Keeps the
surrounding meta-comments, replaces its inside.
*/

webpage.prototype.replace_editable_inside=function(id,value) {
	var atpos=0;
	var ret=[];

	var wholeTag=this.get_editable_inside_pos(id);
	ret.push(this.page.substr(atpos,wholeTag[0]));
	ret.push(value);
	ret.push(this.page.substr(atpos+wholeTag[0]+wholeTag[1]));
	this.page=ret.join("");
}

/*
====get_editable_inside(id)====

Returns a subpage containing only the innards if the editable area with 'id''.
*/

webpage.prototype.get_editable_inside=function(id) {
	var pos=this.get_editable_inside_pos(id);
	return new webpage(this.page.substr(pos[0],pos[1]));
}

/*
====get_editable_outside(id)====

Returns a subpage containing the whole editable area with 'id'', including surrounding meta-comments.
*/

webpage.prototype.get_editable_inside=function(id) {
	var pos=this.get_editable_inside_pos(id);
	return new webpage(this.page.substr(pos[0],pos[1]));
}

/*
====get_editable_outside_pos(String id)====

Returns position and length of editable area with the given id.

*/

webpage.prototype.get_editable_outside_pos=function(id) {
	var startTag='<!-- #Begin';
	var endTag='<!-- #End';

	var pos=0;
	for (;;) {
		var pos=this.page.indexOf(startTag,pos);
		if (pos==-1) throw("webpage: Cannot find editable area '"+id+"'");
		var idStart=this.page.indexOf('"',pos);
		var idStop=this.page.indexOf('"',idStart+1);
		if (this.page.substr(idStart+1,idStop-idStart-1)==id) break;
		pos=idStop;
	}

	var level=1; // at beginpos
	var atPos=pos;

// 1 2 3 2 3 2 1 0
// ( ( ( ) ( ) ) )

	var beginPos=this.page.indexOf(startTag,atPos+1);
	var endPos=this.page.indexOf(endTag,atPos+1);

	for(;;) {
		if (endPos==-1) throw("webpage: Malformed HTML (4)");
		if (beginPos<endPos && beginPos!=-1) {
			level++;
			atPos=beginPos;			
			beginPos=this.page.indexOf(startTag,atPos+1);
		} else {
			level--;
			if (level==0) break;
			atPos=endPos;
			endPos=this.page.indexOf(endTag,atPos+1);
		}
	}
	endPos=this.page.indexOf('>',endPos);
	return [pos,endPos+1-pos];
}

/*
====get_editable_inside_pos(String id)====

Returns position and length of the editable area with the given id.

Rather
sloppy implementation, which does not allow other tags of the same kind to be
nested inside the given tag. Also, requires tag to have an end tag and also does not
accept xml-style empty tags: <br/>.
*/

webpage.prototype.get_editable_inside_pos=function(pos) {
	var ret=this.get_editable_outside_pos(pos);
	var endPos=this.page.lastIndexOf("<",ret[0]+ret[1]-1);
	var startpos=this.page.indexOf('>',ret[0])+1;
	return [startpos,endPos-startpos];
}

