#include "pointer.h"
#include "callc.h"
#include "type.h"
#include "struct.h"
#include <stdarg.h>
#include <stdio.h>


JSBool JS_DLL_CALLBACK JSX_InitContext(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
	// This function is defined in the main executable:

	JSContext *(*InitContext)(JSRuntime *rt);
	JSContext *newcx;
	JSObject *global=0, *array, *pointer;
	struct JSX_pointer *ptr=0;
    jsval tmp;

	InitContext=JSX_globalsym(cx, 0, "JSEXT_InitContext");
	if (!InitContext) return JS_FALSE;

	newcx=(*InitContext)(JS_GetRuntime(cx));
	if (!newcx) return JS_FALSE;
	
    array=JS_NewArrayObject(cx, 2, 0);
	*rval=OBJECT_TO_JSVAL(array);

	pointer=JS_NewObject(cx, &JSX_pointer_class, JSX_pointer_proto, 0);
	tmp=OBJECT_TO_JSVAL(pointer);
    JS_SetElement(cx, array, 0, &tmp);

    ptr=(struct JSX_pointer *)JS_malloc(cx,sizeof(struct JSX_pointer)+sizeof(struct JSX_type)+sizeof(int));
    if (!ptr) return JS_FALSE;

    ptr->ptr=newcx;
    ptr->callback=0;
    ptr->type=(struct JSX_type*)(ptr+1);
	memset(ptr->type,0,sizeof(struct JSX_type));
	ptr->type->Dim=1;
	ptr->type->Array_size=(int *)(ptr->type+1);
	ptr->type->Array_size[0]=0;
	ptr->type->Type=VOIDSYM;
	JS_SetPrivate(cx, pointer, ptr);


	global=JS_GetGlobalObject(newcx);
	tmp=OBJECT_TO_JSVAL(global);
    JS_SetElement(cx, array, 1, &tmp);

	return JS_TRUE;
}


JSBool JSX_ReportException(JSContext *cx, char *format, ...) {
  int len;
  char *msg;
  JSString *Str;
  va_list va;
  jsval str;

  va_start(va, format);
#ifdef WIN32
  msg=JS_malloc(cx, 81);
  va_start(va, format);
  len=_vsnprintf(msg,80,format,va);
  msg[80]=0;
#else
  msg=JS_malloc(cx, 81);
  va_start(va, format);
  len=vsnprintf(msg,80,format,va);
  msg[80]=0;

    /*
  len=vsnprintf(NULL,0,format,va);
  va_end(va);
  msg=JS_malloc(cx, len+1);
  va_start(va, format);
  vsprintf(msg,format,va);
    */
#endif
  va_end(va);
  Str=JS_NewString(cx, msg, len);
  str=STRING_TO_JSVAL(Str);
  JS_SetPendingException(cx, str);

  return JS_FALSE;
}


JS_EXPORT_API(JSBool) JSEXT_init(JSContext *cx, JSObject *obj) {
  JSX_pointer_proto=JS_InitClass(cx, obj, NULL, &JSX_pointer_class, JSX_pointer_new, 1, NULL, NULL, NULL, NULL);
  if (JSX_pointer_proto==0) 
    return JS_FALSE;
  JS_SetPrivate(cx, JSX_pointer_proto, NULL);
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "toNumber", JSX_pointer_toNumber, 1, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "toString", JSX_pointer_toString, 1, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "toSource", JSX_pointer_toString, 1, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "add", JSX_pointer_add, 1, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "apply", JSX_pointer_apply, 1, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "call", JSX_pointer_call, 1, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "get", JSX_pointer_get, 0, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "set", JSX_pointer_set, 1, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "member", JSX_pointer_member, 1, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "string", JSX_pointer_string, 1, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "UCstring", JSX_pointer_UCString, 1, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "sizeof", JSX_pointer_sizeof, 0, 0)) return JS_FALSE;
  if (NULL==JS_DefineFunction (cx, JSX_pointer_proto, "type", JSX_pointer_type, 1, 0)) return JS_FALSE;
 
  JSX_type_proto=JS_InitClass(cx, obj, NULL, &JSX_type_class, JSX_Type_new, 1, NULL, NULL, NULL, NULL);
  if (JSX_type_proto==0) return JS_FALSE;
  JS_SetPrivate(cx, JSX_type_proto, NULL);

  if (NULL==JS_DefineFunction (cx, obj, "InitContext", JSX_InitContext, 0, 0)) return JS_FALSE;

#ifdef JS_THREADSAFE
  JSX_callback_lock=PR_NewLock();  
#endif  

  return JS_TRUE;
}
