/*
* Copyright (c) 2006, Berges Allmenndigitale Rdgivningstjeneste
* All rights reserved.
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of Berges Allmenndigitale Rdgivningstjeneste nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE BERGES AND CONTRIBUTORS ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE BERGES AND CONTRIBUTORS BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "type.h"
#include "pointer.h"
#include "js_to_c2.h"
#include "../js/src/jsfun.h"

#ifdef WIN32
# include <malloc.h>
#else
# include <alloca.h>
#endif

#include "callc.h"
#include <string.h>

/*
  convert an object to a struct, return size of C struct
*/

static size_t JSX_JS_struct(JSContext *cx, JSObject *this, 
		      char *C_var, struct JSX_type *C,
		      JSObject *object, int will_clean ) {

  int i;
  char *pC_var;

  struct JSX_struct *strct=C->Structtype;
  if (!strct) strct=JSX_lookupstruct(cx, this, C->Struct);
  if (!strct) return 0;
  
  if (C && C->Dim) {
	if (will_clean) {
		*(char **) C_var = (char *)calloc(strct->Sizeof,1);
	    pC_var = *(char **)C_var;
	} else return 0;
  }

  for (i=0; i<strct->nMembers; i++) {
    jsval e;
    if (JS_GetProperty(cx, object, strct->Members[i]->Name, &e)==JS_FALSE) return 0;
    if (e!=JSVAL_VOID) {
		size_t sz=
      JSX_JS_to_C(cx, this, pC_var+strct->Offsets[i], strct->Members[i], &e, will_clean, 0);
		if (sz==0) goto failure;
    }
  }

  if (C && C->Dim)
    return sizeof(char *);
  else
    return strct->Sizeof;

failure:
  if (C && C->Dim) free(*(char **)C_var);
  return 0;
}


JSBool JSX_ConvertError(JSContext *cx, struct JSX_type *C, char *msg) {
  if (C && C->Name) {
#ifdef WIN32
    char *tmpmsg=_alloca(strlen(msg+20)+strlen(C->Name));
#else
    char *tmpmsg=alloca(strlen(msg+20)+strlen(C->Name));
#endif
    sprintf(tmpmsg,"%s in %s",msg, C->Name);
    JSX_ReportException(cx, tmpmsg);
  } else {
    JSX_ReportException(cx, msg);
  }
  return JS_FALSE;
}

size_t JSX_JS_to_C_multi(JSContext * cx, JSObject * this, char **argbuf,
			  struct JSX_type **C, int nArgs_C, jsval *v, int nArgs_v, int will_clean, int align) {

  int i;
  size_t arg_size;
  size_t fixarray_size;
  char *C_var;
  char *fixarrays;

  if (will_clean) {
    arg_size=0;
    fixarray_size=0;
    for (i=0; i<nArgs_v; i++) {
      struct JSX_type *Ci=i<nArgs_C?C[i]:0;
      int sz;
      if (Ci && Ci->Dim && *Ci->Array_size) {
	fixarray_size+=JSX_sizeof(cx, this, Ci, v+i);
	sz=sizeof(char *);
      } else {
	sz=JSX_sizeof(cx, this, Ci, v+i);
      }

      arg_size+=align*((sz+align-1)/align);
    }
    
    if (arg_size==0) {
      *argbuf=0;//JS_malloc(cx, 1);
      return 0;
    }
    
    *argbuf=JS_malloc(cx, arg_size+fixarray_size);
    if (!*argbuf) {
      JSX_ReportException(cx, "Allocation error");
      return 0;
    }
  }

  C_var=*argbuf;
  fixarrays=C_var+arg_size;
  
  for (i=0; i<nArgs_v; i++) {
    size_t elem_size;
    struct JSX_type *Ci=i<nArgs_C?C[i]:0;
    
    if (Ci && Ci->Dim && *Ci->Array_size) {
      // Array[n] are an exception in that they are passed by pointer to functions
      // but stored inline in structs.
      if (!will_clean) goto failure;
      *(char **)C_var=fixarrays;
      elem_size=JSX_JS_to_C(cx, this, fixarrays, Ci, v+i, will_clean, 0);
	  if (elem_size) {
	      fixarrays+=elem_size;
		  elem_size=sizeof(char *);
	  }
    } else {
      elem_size=JSX_JS_to_C(cx, this, C_var, Ci, v+i, will_clean, 0);
    }

    if (!elem_size) goto failure;
    C_var+=align*((elem_size+align-1)/align);
  }

  return C_var-*argbuf;

 failure:
  if (will_clean && *argbuf) {
	  JS_free(cx,*argbuf);
	  *argbuf=0;
  }
  return 0;
}

size_t
JSX_known_array_to_C (JSContext * cx, JSObject * this,
		       char *C_var, struct JSX_type *C,
		       JSObject * array, int will_clean, char *oldptr)
{

  int arraylength;
  int i;
  size_t buf_size = 0;
  struct JSX_type C2, *C2p;

  JS_GetArrayLength (cx, array, &arraylength);
  if (C) {
    memcpy(&C2,C, sizeof(C2));
    C2.Array_size++;
    C2.Dim--;
    C2p=&C2;
  } else {
    C2p=0;
  }

  if (JSX_contains_pointers(C) && oldptr)
    {
      // Contains pointers
      
      for (i = 0; i < arraylength; i++)
	{
	  jsval tmp;
	  size_t elem_size;

		  JS_GetElement (cx, array, i, &tmp);
	  elem_size = JSX_JS_to_C (cx, this, C_var, C2p,
					   &tmp, will_clean, oldptr);
	  if (!elem_size)
	    return 0;

	  C_var += elem_size;
	  oldptr += elem_size;
	  buf_size+=elem_size;
	}
    }
  else
    {

      // Does not contain pointers

      for (i = 0; i < arraylength; i++)
	{
	  jsval tmp;
	  size_t elem_size;

	  JS_GetElement (cx, array, i, &tmp);
	  elem_size = JSX_JS_to_C (cx, this, C_var, C2p,
					   &tmp, will_clean, oldptr);
	  if (!elem_size)
	    return 0;
	  
	  C_var += elem_size;
	  buf_size+=elem_size;
	}
    }
  return buf_size;
}


size_t
JSX_unknown_array_to_C (JSContext * cx, JSObject * this,
			 char *C_var,
			 JSObject * array, int will_clean, char *oldptr)
{
  int arraylength;
  int i;
  size_t buf_size = 0;

  JS_GetArrayLength (cx, array, &arraylength);

  for (i = 0; i < arraylength; i++)
    {
      jsval tmp;
      size_t elem_size;
      JS_GetElement (cx, array, i, &tmp);

      if (!JSVAL_IS_NULL(tmp) && JSVAL_IS_OBJECT (tmp) &&
	  JS_IsArrayObject (cx, JSVAL_TO_OBJECT (tmp)))
	{

	  // Contains pointer

	  elem_size = JSX_JS_to_C (cx, this, C_var, 0,
				    &tmp, will_clean, oldptr);
	  if (!elem_size)
	    return 0;

	  oldptr += elem_size;
	}
      else
	{

	  // Does not contain pointer

	  elem_size = JSX_JS_to_C (cx, this, C_var, 0,
				    &tmp, will_clean, oldptr);
	  if (!elem_size)
	    return 0;

	}
      C_var += elem_size;
      buf_size+=elem_size;
    }
  return buf_size;
}


size_t
JSX_JS_to_C (JSContext * cx, JSObject * this,
	      char *C_var, struct JSX_type *C,
	      jsval * v, int will_clean, char *oldptr)
{

  JSObject *object;

  if (*v==JSVAL_VOID) {
    return JSX_sizeof (cx, this, C, v);
  }

  if (!JSVAL_IS_NULL(*v) && JSVAL_IS_OBJECT (*v))
    {
      object = JSVAL_TO_OBJECT (*v);
    }
  else
    {
      object = 0;
    }

  if ((!C || (C->Type==VASYM && C->Dim==0)  || (C->Dim==1 && C->Array_size[0]==0 && C->Type==VOIDSYM)) &&
	  object && JS_IsArrayObject (cx, object) && will_clean)
    {
      // JS: Array
      // C: unknown, va or void *

      JSObject *array = object;
      int arraylength;
      size_t buf_size = 0;
      size_t extra_size=0;
      int i;
      size_t ret;

      JS_GetArrayLength (cx, array, &arraylength);

      if (!arraylength)
	{			// empty array -> null pointer
	  *(void **) C_var = 0;
	  return JS_TRUE;
	}

      for (i = 0; i < arraylength; i++)
	{
	  jsval tmp;
	  size_t elem_size;

	  JS_GetElement (cx, array, i, &tmp);

	  elem_size = JSX_sizeof (cx, this, 0, &tmp);
	  buf_size += elem_size;
	  if (!JSVAL_IS_NULL(tmp) && JSVAL_IS_OBJECT (tmp) &&
	      JS_IsArrayObject (cx, JSVAL_TO_OBJECT (tmp)))
	    {
	      // Allocate room for backup pointer
	      extra_size += elem_size;
	    }
	}

      *(char **) C_var = JS_malloc (cx, buf_size+extra_size);
      if (oldptr) *(char **)oldptr=*(char **)C_var;

      ret = JSX_unknown_array_to_C (cx, this,
					    *(char **) C_var,
					    array, will_clean,
					    *(char **) C_var+buf_size);
      if (!ret)
	{
	  JS_free (cx, *(char **) C_var);
	  return JS_FALSE;
	}
      
      return sizeof(char *);
    }

  if (!C || C->Dim == 0)
    {
      switch (C?C->Type:0)
	{

	case 0:		// C->Type==unknown, C->Dim==0
	  if (JSVAL_IS_NULL (*v))
	    {
	      *(void **) C_var = 0;
	      return sizeof (void *);
	    }
	  if (object &&
	      JS_InstanceOf (cx, object, &JSX_pointer_class, 0))
	    {
	      struct JSX_pointer *priv= JS_GetPrivate (cx, object);
	      if (!priv) return 0;
	      *(void **) C_var = priv->ptr;
	      return sizeof (void *);
	    }
	  if (JSVAL_IS_STRING (*v))
	    {
	      *(char **) C_var = JS_GetStringBytes (JSVAL_TO_STRING (*v));
	      return sizeof (char *);
	    }
	  if (JSVAL_IS_BOOLEAN (*v))
	    {
	      *(int *) C_var = JSVAL_TO_BOOLEAN (*v);
	      return sizeof (int);
	    }
	  if (JSVAL_IS_INT (*v))
	    {
	      *(int *) C_var = JSVAL_TO_INT (*v);
	      return sizeof (int);
	    }
	  if (JSVAL_IS_NUMBER (*v))
	    {
	      JS_ValueToNumber (cx, *v, (jsdouble *) C_var);
	      return sizeof (jsdouble);
	    }
	  break;

	case CHARSYM:		// C->Type==CHAR, C->Dim==0
	  if (JSVAL_IS_INT (*v))
	    {
	      *(char *) C_var = (char) JSVAL_TO_INT (*v);
	      return sizeof (char);
	    }
	  if (JSVAL_IS_BOOLEAN (*v))
	    {
	      *(char *) C_var = (char) JSVAL_TO_BOOLEAN (*v);
	      return sizeof (char);
	    }
	  if (JSVAL_IS_NUMBER (*v))
	    {
	      jsdouble dblbuf;
	      JS_ValueToNumber (cx, *v, &dblbuf);
	      *(char *) C_var = (char) dblbuf;
	      return sizeof (char);
	    }
	  {
	    JSString *str = JS_ValueToString (cx, *v);
	    if (JS_GetStringLength (str) < 1)
	      {
		*(char *) C_var = 0;
	      }
	    else
	      {
		char *cstr = JS_GetStringBytes (str);
		*(char *) C_var = cstr[0];
	      }
	    return sizeof (char);
	  }

	case SHORTSYM:		// C->Type==SHORT, C->Dim==0
	  if (JSVAL_IS_INT (*v))
	    {
	      *(int16 *) C_var = (int16) JSVAL_TO_INT (*v);
	      return sizeof (int16);
	    }
	  if (JSVAL_IS_BOOLEAN (*v))
	    {
	      *(int16 *) C_var = (int16) JSVAL_TO_BOOLEAN (*v);
	      return sizeof (int16);
	    }
	  if (JSVAL_IS_NUMBER (*v))
	    {
	      jsdouble dblbuf;
	      JS_ValueToNumber (cx, *v, &dblbuf);
	      *(int16 *) C_var = (int16) dblbuf;
	      return sizeof (int16);
	    }
	  {
	    JSString *str = JS_ValueToString (cx, *v);
	    if (JS_GetStringLength (str) < 1)
	      {
		*(jschar *) C_var = 0;
	      }
	    else
	      {
		jschar *cstr = JS_GetStringChars (str);
		*(jschar *) C_var = cstr[0];
	      }
	  }
	  return sizeof (jschar);

	case IFSYM:		// C->Type=boolean, C->Dim=0
	  {
	    JSBool buf;
	    if (JS_FALSE == JS_ValueToBoolean (cx, *v, &buf))
	      {
		JSX_ConvertError(cx, C, "Type mismatch, expected boolean");
		return 0;
	      }
	    switch (C->Size)
	      {
	      case 0:		// unknown type
		*(int *) C_var = (int) buf;
		return sizeof (int);
	      case 1:
		*(char *) C_var = (char) buf;
		return sizeof (int);
	      case 2:
		*(int16 *) C_var = (int16) buf;
		return sizeof (int16);
	      case 4:
		*(int32 *) C_var = (int32) buf;
		return sizeof (int32);
	      default:
		JSX_ConvertError(cx, C, "No support for given int size");
		return 0;
	      }
	  break;

	case INTSYM:		// C->Type=int, C->Dim=0
	    {
		int32 buf;
	      
		  if (object &&
		     JS_InstanceOf (cx, object, &JSX_pointer_class, 0))
		  {
			  struct JSX_pointer *priv= JS_GetPrivate (cx, object);
			  if (!priv) return 0;
			  *(void **) C_var = priv->ptr;
			  return sizeof (void *);

		  } else if (JSVAL_IS_INT(*v)) {
			if (JS_FALSE == JS_ValueToInt32 (cx, *v, &buf)) {
				JSX_ConvertError(cx, C, "Type mismatch, expected integer");
			    return 0;
			}
		  } else {
			jsdouble dbuf;
			if (JS_FALSE == JS_ValueToNumber (cx, *v, &dbuf)) {
			  JSX_ConvertError(cx, C, "Type mismatch, expected integer");
			  return 0;
			}
			buf=(int32)dbuf;
		  }

	      switch (C->Size)
		{
		case 0:	// unknown type
		  *(int *) C_var = (int) buf;
		  return sizeof (int);
		case 1:
		  *(char *) C_var = (char) buf;
		  return sizeof (int);
		case 2:
		  *(int16 *) C_var = (int16) buf;
		  return sizeof (int16);
		case 4:
		  *(int32 *) C_var = (int32) buf;
		  return sizeof (int32);
		default:
		  JSX_ConvertError(cx, C, "No support for given int size");
		  return 0;
		}
	    }

	case DOUBLESYM:		// C->Type=double, C->Dim=0
	    {
	      jsdouble buf;
	      if (JS_FALSE == JS_ValueToNumber (cx, *v, &buf))
		{
		  JSX_ConvertError(cx, C, "Type mismatch, expected number");
		  return 0;
		}
	      switch (C->Size)
		{
		case 0:	// unknown type
		  *(double *) C_var = (double) buf;
		  return sizeof (int);
		case sizeof (float):
		  *(float *) C_var = (float) buf;
		  return sizeof (float);
		case sizeof (double):
		  *(double *) C_var = (double) buf;
		  return sizeof (double);
		default:
		  JSX_ConvertError(cx, C, "No support for given float size");
		  return 0;
		}
	    }
	  }
	}
      JSX_ConvertError(cx, C, "Unknown type");
      return 0;
    }

  if (C && C->Dim == 1 && JSVAL_IS_STRING (*v))
    {
      switch (C->Type)
	{
	case CHARSYM:
	case VOIDSYM:
	  {
	    JSString *str = JS_ValueToString (cx, *v);

	  if (C->Array_size[0]) {
		// JS: String
		// C: char [n]

		char *strb=JS_GetStringBytes(str);
		int len=JS_GetStringLength(str);
		if (len>=C->Array_size[0]) {
			memcpy(C_var, strb, C->Array_size[0]);
		} else {
			memcpy(C_var, strb, len);
			memset(C_var+len, 0, C->Array_size[0]-len);
		}
		return sizeof(char)*C->Array_size[0];
	  } else {
		// JS: String
		// C: char*

		*(char **) C_var = JS_GetStringBytes (str);
	    return sizeof (char *);
	  }
	  }
	case SHORTSYM:
	  {
	    JSString *str = JS_ValueToString (cx, *v);

	  if (C->Array_size[0] && C->Type==SHORTSYM) {
		// JS: String
		// C: short [n]

		jschar *strc=JS_GetStringChars(str);
		int len=JS_GetStringLength(str);
		if (len>=C->Array_size[0]) {
			memcpy(C_var, strc, C->Array_size[0]*sizeof(jschar));
		} else {
			memcpy(C_var, strc, len*sizeof(jschar));
			memset(C_var+len*sizeof(jschar), 0, (C->Array_size[0]-len)*sizeof(jschar));
		}
		return sizeof(jschar)*C->Array_size[0];
	  } else {
		// JS: String
		// C: short*
	
		*(jschar **) C_var = JS_GetStringChars (str);
	    return sizeof (jschar *);
	  }
	  }
	default:
	  // should never get here, logically
	  JSX_ConvertError(cx, C, "Type mismatch, expected string");
	  return 0;
	}
    }

  if (C && C->Dim==0 && (C->Type==STRUCTSYM || C->Type==UNIONSYM) && object)
    { // C: Direct struct
      // JS: Object
      return JSX_JS_struct(cx, this, C_var, C, object, will_clean);
    }
  
  if (C && C->Dim == 1 && *C->Array_size == 0) {
	  if (JSVAL_IS_INT(*v)) { 
		  // C: *
		  // JS: int
		*(int *)C_var=JSVAL_TO_INT(*v);
		return sizeof(int);
	  }

	  if (JSVAL_IS_DOUBLE(*v)) { 
		  // C: *
		  // JS: number
		jsdouble tmp;
		JS_ValueToNumber(cx, *v, &tmp);
		*(int *)C_var=(unsigned int)tmp;
		return sizeof(int);
	  }
	}

  if (C && C->Dim >= 1 && *C->Array_size == 0)
    { // C: *
      // JS: Any
      if (JSVAL_IS_NULL (*v))
	{
	  *(void **) C_var = 0;
	  return sizeof (void *);
	}
      if (object &&
	  JS_InstanceOf (cx, object, &JSX_pointer_class, 0))
	{
	  struct JSX_pointer *priv= JS_GetPrivate (cx, object);
	  if (!priv) return 0;
	  *(void **) C_var = priv->ptr;
	  if (priv->callback) { // Callback function, copy function type back to js pointer
	    priv=JSX_pointer_settype(cx, this, priv, C);
	    JS_SetPrivate(cx, object, priv);
	  }
	  return sizeof (void *);
	}

      if (C->Dim == 1 && *C->Array_size==0 && object && will_clean &&
	  (C->Type==STRUCTSYM || C->Type==UNIONSYM))
	{
	  // C: Struct *
	  // JS: Object
	  if (JSX_JS_struct(cx, this, C_var, C, object, will_clean))
	    return sizeof(char *);
	  else
	    return 0;
	}
      
  }
  
  if (C && C->Dim > 0 && *C->Array_size)
    {
      // Convert [n][m][y] arrays
      // implicitly C->Type && C->Size are def'd

      int arraylength;
      JSObject *array = object;


      if (!array || !JS_IsArrayObject (cx, array))
	{
	  JSX_ConvertError(cx, C, "Type mismatch, expected array");
	  return 0;
	}
      
      JS_GetArrayLength (cx, array, &arraylength);
      if (arraylength != *C->Array_size)
	{
	  JSX_ConvertError(cx, C, "Wrong array length");
	  return 0;
	}
      
      if (JSX_contains_pointers(C)) {
	
	// Contains pointers
	
	if (!oldptr) oldptr=C_var+JSX_sizeof (cx, this, C, 0);
	
      }
      
      return JSX_known_array_to_C (cx, this,
				    C_var, C,
				    array, will_clean, 
				    oldptr);
    }


  if (C && object && JS_IsArrayObject (cx, object) && C->Dim > 0 && will_clean
      && *C->Array_size == 0 && (C->Dim > 1
				|| (C->Type != VOIDSYM
				    && C->Type != RETURNSYM)))
    {

      // Rollout known * arrays

      JSObject *array = object;
      int arraylength=0;
      size_t elem_size;
      size_t ret;
      struct JSX_type C2, *C2p;

      if (JS_IsArrayObject(cx, object)) JS_GetArrayLength (cx, array, &arraylength);

      if (!arraylength)
	{			// empty array -> null pointer
	  *(void **) C_var = 0;
	  return sizeof(void *);//JS_TRUE;
	}

      if (C) {
	memcpy(&C2,C, sizeof(C2));
	C2.Array_size++;
	C2.Dim--;
	C2p=&C2;
      } else {
	C2p=0;
      }

	  elem_size = JSX_sizeof (cx, this, C2p, 0);

      if (JSX_contains_pointers(C) && !oldptr)
	{

	  // Contains pointers, no space has been allocated.

	  *(char **) C_var = JS_malloc (cx, elem_size * arraylength * 2);
	  oldptr=*(char **) C_var + elem_size * arraylength;
	  
	}
      else
	{
	  
	  // Does not contain pointers
	  
	  *(char **) C_var = JS_malloc (cx, elem_size * arraylength);
	  if (oldptr) *(char **)oldptr=*(char **)C_var;
	  
	}

      ret= JSX_known_array_to_C (cx, this,
				  *(char **) C_var, C,
				  array, will_clean, oldptr);

      if (!ret)
	{
	  JS_free (cx, *(char **) C_var);
	  return JS_FALSE;
	}

      return sizeof(void *);//elem_size * arraylength;
    }

  JSX_ConvertError(cx, C, "Type mismatch (3)");
  return 0;

}
