#include "type.h"
#include "struct.h"
#include "pointer.h"
#include "callc.h"

#include <string.h>

/*
  Find length of string representation of a JSX_functype
*/

int JSX_functype_string_len(JSContext *cx, struct JSX_functype *Func) {
  int ret=1; // callconv
  char tmp[10];
  int i;

  if (Func->Ret) {
    ret+=JSX_type_string_len(cx, Func->Ret);
  } else {
    ret++;
  }

  ret+=sprintf(tmp,"%d ",Func->nArgs);
  for (i=0; i<Func->nArgs; i++) {
    ret+=JSX_type_string_len(cx, Func->Args[i]);
  }
  ret++;
  return ret;
}

/*
  Find length of string representation of a JSX_struct
*/

int JSX_struct_string_len(JSContext *cx, struct JSX_struct *strct) {
  int ret=0;
  char tmp[10];
  int i;

  ret+=sprintf(tmp,"%d ",strct->nMembers);
  for (i=0; i<strct->nMembers; i++) {
    ret+=JSX_type_string_len(cx, strct->Members[i]);
  }
  return ret;
}

/*
  Find length of string representation of a JSX_type
*/

int JSX_type_string_len(JSContext *cx, struct JSX_type *type) {
  int ret=0;
  char tmp[10];
  int i;

  ret+=strlen(type->Name)+1;
  ret+=sprintf(tmp,"%d ",type->Dim);
  for (i=0; i<type->Dim; i++)
    ret+=sprintf(tmp,"%d ",type->Array_size[i]);

  switch(type->Type) {
  case STRUCTSYM:
    if (type->Structtype) {
      ret++;
      ret+=JSX_struct_string_len(cx, type->Structtype);
    } else {
      ret++;
      ret+=strlen(type->Struct)+1;
      break;
    }
  case UNIONSYM:
    if (type->Structtype) {
      ret++;
      ret+=JSX_struct_string_len(cx, type->Structtype);
    } else {
      ret++;
      ret+=strlen(type->Struct)+1;
      break;
    }
  case DOUBLESYM:
    ret+=2;
    break;
  case CHARSYM:
  case SHORTSYM:
  case INTSYM:
    ret+=3;
    break;
  case VASYM:
  case VOIDSYM:
    ret++;
    break;
  case RETURNSYM:
    ret++;
    ret+=JSX_functype_string_len(cx, type->Func);
    break;
  }

  return ret;
}

/*
  Make a string representation of a JSX_functype
*/

char *JSX_functype_string(JSContext *cx, struct JSX_functype *Func, char *t) {
  int i;

  *(t++)=(Func->Callconv==CDECLSYM)?'c':'s';

  if (Func->Ret) {
    t=JSX_type_string(cx, Func->Ret, t);
  } else {
    *(t++)='v';
  }

  t+=sprintf(t,"%d ",Func->nArgs);
  for (i=0; i<Func->nArgs; i++) {
    t=JSX_type_string(cx, Func->Args[i], t);
  }
  *(t++)=Func->Elipsis+'0';
  return t;
}

/*
  Make a string representation of a JSX_struct
*/

char *JSX_struct_string(JSContext *cx, struct JSX_struct *strct, char *t) {
  int i;

  t+=sprintf(t,"%d ",strct->nMembers);
  for (i=0; i<strct->nMembers; i++) {
    t=JSX_type_string(cx, strct->Members[i], t);
  }
  return t;
}

/*
  Make a string representation of a JSX_type
*/

char *JSX_type_string(JSContext *cx, struct JSX_type *type, char *t) {
  int i;

  t+=sprintf(t,"%s %d ",type->Name,type->Dim);
  for (i=0; i<type->Dim; i++)
    t+=sprintf(t,"%d ",type->Array_size[i]);

  switch(type->Type) {
  case STRUCTSYM:
    if (type->Structtype) {
      *(t++)='S';
      t=JSX_struct_string(cx, type->Structtype, t);
    } else {
      *(t++)='s';
      t+=sprintf(t,"%s ",type->Struct);
      break;
    }
  case UNIONSYM:
    if (type->Structtype) {
      *(t++)='U';
      t=JSX_struct_string(cx, type->Structtype, t);
    } else {
      *(t++)='u';
      t+=sprintf(t,"%s ",type->Struct);
      break;
    }
  case DOUBLESYM:
    *(t++)='f';
    if (type->Size==sizeof(float)) *(t++)='0';
    else if (type->Size==sizeof(double)) *(t++)='1';
    break;
  case CHARSYM:
    *(t++)='i';
    *(t++)='0';
    break;
  case SHORTSYM:
    *(t++)='i';
    *(t++)='1';
    break;
  case INTSYM:
    *(t++)='i';
    if (type->Size==sizeof(int)) *(t++)='2';
    else if (type->Size==sizeof(long)) *(t++)='3';
    break;
  case VASYM:
    *(t++)='V';
    break;
  case VOIDSYM:
    *(t++)='v';
    break;
  case RETURNSYM:
    *(t++)='F';
    t=JSX_functype_string(cx, type->Func, t);
    break;
  }

  switch(type->Type) {
  case CHARSYM:
  case SHORTSYM:
  case INTSYM:
    switch(type->Signed) {
    case 0:
      *(t++)='0';
      break;
    case 1:
      *(t++)='1';
      break;
    case 2:
      *(t++)='?';
      break;
    }
  }

  return t;
}

/*
  Copy a struct JSX_functype
 */


void JSX_functype_copy(JSContext *cx, struct JSX_functype *F, char **tb, char **tt) {
  struct JSX_functype *ret;
  int i;

  if (!F) return;
  ret=(struct JSX_functype *)*tb;
  (*tb)+=sizeof(struct JSX_functype);

  ret->Callconv=F->Callconv;
  ret->nArgs=F->nArgs;
  ret->Elipsis=F->Elipsis;
  ret->Args=(struct JSX_type **)*tb;
  (*tb)+=sizeof(struct JSX_type *)*F->nArgs;

  for (i=0; i<F->nArgs; i++) {
    ret->Args[i]=(struct JSX_type *)*tb;
    JSX_type_copy(cx, F->Args[i], tb, tt);
  }

  if (F->Ret) {
    ret->Ret=(struct JSX_type *)*tb;
    JSX_type_copy(cx, F->Ret, tb, tt);
  } else {
    ret->Ret=0;
  }
}


/*
  Copy a struct JSX_struct
 */

void JSX_struct_copy(JSContext *cx, struct JSX_struct *C, char **tb, char **tt) {
  struct JSX_struct *ret;
  int i;

  if (!C) return;
  ret=(struct JSX_struct *)*tb;
  (*tb)+=sizeof(struct JSX_struct);
  ret->nMembers=C->nMembers;
  ret->Members=(struct JSX_type **)*tb;
  (*tb)+=sizeof(struct JSX_type *)*C->nMembers;

  ret->Offsets=(int *)*tb;
  (*tb)+=sizeof(int)*C->nMembers;
  
  for (i=0; i<C->nMembers; i++) {
    ret->Members[i]=(struct JSX_type *)*tb;
    JSX_type_copy(cx, C->Members[i], tb, tt);
  }

  memcpy(ret->Offsets, C->Offsets, sizeof(int)*C->nMembers);
}

/*
  Copy a struct JSX_type
 */

void JSX_type_copy(JSContext *cx, struct JSX_type *C, char **tb, char **tt) {
  struct JSX_type *ret;

  if (!C) return;

  ret=(struct JSX_type *)*tb;
  (*tb)+=sizeof(struct JSX_type);
  ret->Type=C->Type;
  ret->Size=C->Size;
  ret->Dim=C->Dim;
  ret->Signed=C->Signed;
  ret->Const=C->Const;
  ret->dl=C->dl;

  if (C->Func) {
    ret->Func=(struct JSX_functype *)*tb;
    JSX_functype_copy(cx, C->Func, tb, tt);
  } else {
    ret->Func=0;
  }

  if (C->Structtype) {
    ret->Structtype=(struct JSX_struct *)*tb;
    JSX_struct_copy(cx, C->Structtype, tb, tt);
  } else {
    ret->Structtype=0;    
  }
  
  ret->Array_size=(int *)*tb;
  if (C->Dim) {
    (*tb)+=sizeof(int)*C->Dim;
    memcpy(ret->Array_size,C->Array_size,sizeof(int)*C->Dim);
  }

  if (C->Struct) {
    *tt-=strlen(C->Struct)+1;
    ret->Struct=*tt;
    strcpy(ret->Struct, C->Struct);
  } else {
    ret->Struct=0;
  }

  if (C->Name) {
    *tt-=strlen(C->Name)+1;
    ret->Name=*tt;
    strcpy(ret->Name, C->Name);
  } else {
    ret->Name=0;
  }
}

int JSX_functype_size(struct JSX_functype *F) {
  int size;
  int i;

  if (F==0) return 0;
  size=sizeof(struct JSX_functype);
  size+=sizeof(struct JSX_type *)*F->nArgs;
  if (F->Ret) size+=JSX_type_size(F->Ret);
  for (i=0; i<F->nArgs; i++)
    size+=JSX_type_size(F->Args[i]);
  return size;
}

int JSX_struct_size(struct JSX_struct *S) {
  int i;
  int size;

  if (S==0) return 0;
  size=sizeof(struct JSX_struct);
  size+=sizeof(int)*S->nMembers;
  size+=sizeof(struct JSX_type *)*S->nMembers;
  for (i=0; i<S->nMembers; i++)
    size+=JSX_type_size(S->Members[i]);
  return size;
}

int JSX_type_size(struct JSX_type *C) {
  int size;

  if (C==0) return 0;
  size=sizeof(struct JSX_type);
  size+=sizeof(int)*C->Dim;
  if (C->Struct) size+=strlen(C->Struct)+1;
  if (C->Func) size+=JSX_functype_size(C->Func);
  if (C->Name) size+=strlen(C->Name)+1;
  if (C->Structtype) size+=JSX_struct_size(C->Structtype);
  return size;
}

int JSX_type_size_s(char **_s) {
  char *s=*_s;
  int size=sizeof(struct JSX_type);
  char *aftername=strchr(s,' ');
  int dim;
  int i;
  int tmp;
  char type;

  if (!aftername) return -1;
  size+=aftername-s+1;
  s=aftername;
  if (!*s) return -1;
  s++;
  dim=atoi(s);
  s=strchr(s,' ');
  if (!s) return -1;
  s++;
  size+=sizeof(int)*dim;
  for (i=0; i<dim; i++) {
    s=strchr(s,' ');
    if (!s) return -1;
    s++;
  }
  type=*(s++);
  
  switch(type) {
  case 's': // struct
  case 'u': // union
    aftername=strchr(s,' ');
    if (!aftername) return -1;
    size+=aftername-s+1;
    s=aftername+1;
    break;
  case 'f':
    s++; // varsize;
    break;
  case 'i':
    s++; // varsize;
    if (!*s) return -1;
    s++; // signed;
    break;
  case 'V':
    break;
  case 'v':
    break;
  case 'S':
  case 'U':
    s--;
    tmp=JSX_struct_size_s(&s);
    if (tmp==-1) return -1;
    size+=tmp;
    break;
  case 'F':
    s--;
    tmp=JSX_functype_size_s(&s);
    if (tmp==-1) return -1;
    size+=tmp;
    break;
  default:
    return -1;
  }

  *_s=s;
  return size;
}


int JSX_functype_size_s(char **_s) {
  char *s=*_s;
  int size=sizeof(struct JSX_functype);
  int tmp;
  int dim;
  int i;

  s++; // 'F'
  s++; // callconv
  if (*s!='v') {
    tmp=JSX_type_size_s(&s); // return type
    if (tmp==-1) return -1;
    size+=tmp;
  } else {
    s++;
  }
  dim=atoi(s);
  s=strchr(s,' ');
  if (!s) return -1;
  s++;
  size+=sizeof(struct JSX_type *)*dim;
  for (i=0; i<dim; i++) {
    s++; // const
    size+=JSX_type_size_s(&s); // param
  }
  if (!*s) return -1;
  s++; // elipsis
  *_s=s;
  return size;
}


int JSX_struct_size_s(char **_s) {
  char *s=*_s;
  int ret=sizeof(struct JSX_struct);
  int nMembers;
  int i;
  int tmp;

  s++; // S or U
  nMembers=atoi(s);
  s=strchr(s,' ');
  if (!s) return -1;
  s++;

  ret+=sizeof(struct JSX_type *)*nMembers;
  ret+=sizeof(int)*nMembers;

  for (i=0; i<nMembers; i++) {
    tmp=JSX_type_size_s(&s);
    if (tmp==-1) return -1;
    ret+=tmp;
  }
  *_s=s;
  return ret;
}

void JSX_struct_new(JSContext *cx, JSObject *obj, char **_s, char **tb, char **tt) {
  char *s=*_s;
  struct JSX_struct *strct=(struct JSX_struct *)*tb;
  int isUnion;
  int i;

  memset(strct, 0, sizeof(struct JSX_struct));
  (*tb)+=sizeof(struct JSX_struct);

  isUnion=(*s=='U');
  s++;

  strct->nMembers=atoi(s);
  s=strchr(s,' ')+1;

  strct->Members=(struct JSX_type **)*tb;
  (*tb)+=sizeof(struct JSX_type *)*strct->nMembers;

  strct->Offsets=(int *)*tb;
  (*tb)+=sizeof(int)*strct->nMembers;

  strct->Sizeof=0;

  for (i=0; i<strct->nMembers; i++) {
    size_t thisalign;
    strct->Members[i]=(struct JSX_type *)*tb;
    JSX_type_new(cx, obj, &s, tb, tt);
    thisalign=JSX_align(cx, obj, strct->Members[i], 0);
    strct->Sizeof+=(thisalign - strct->Sizeof % thisalign) % thisalign;
    strct->Offsets[i]=isUnion?0:strct->Sizeof;
    strct->Sizeof+=JSX_sizeof(cx, obj, strct->Members[i], 0);
  }
  *_s=s;
}

void JSX_type_new(JSContext *cx, JSObject *obj, char **_s, char **tb, char **tt) {
  char *s=*_s;
  struct JSX_type *type=(struct JSX_type *)*tb;
  char *aftername;
  int i;
  char typechar;

  memset(type, 0, sizeof(struct JSX_type));
  (*tb)+=sizeof(struct JSX_type);
  aftername=strchr(s,' ');
  (*tt)-=aftername-s+1;
  memcpy(*tt,s,aftername-s);
  type->Name=*tt;
  type->Name[aftername-s]=0;
  s=aftername+1;
  type->Dim=atoi(s);
  s=strchr(s,' ')+1;
  type->Array_size=(int *)*tb;
  (*tb)+=sizeof(int)*type->Dim;
  for (i=0; i<type->Dim; i++) {
    type->Array_size[i]=atoi(s);
    s=strchr(s,' ')+1;
  }
  typechar=*(s++);
  
  switch(typechar) {
  case 's': // struct
    type->Type=STRUCTSYM;
    aftername=strchr(s,' ');
    (*tt)-=aftername-s+1;
    memcpy(*tt,s,aftername-s);
    type->Struct=*tt;
    type->Struct[aftername-s]=0;
    s=aftername+1;
    break;
  case 'u': // union
    type->Type=UNIONSYM;
    aftername=strchr(s,' ');
    (*tt)-=aftername-s+1;
    memcpy(*tt,s,aftername-s);
    type->Struct=*tt;
    type->Struct[aftername-s]=0;
    s=aftername+1;
    break;
  case 'f':
    type->Type=DOUBLESYM;
    switch(*s) {
    case '0':
      type->Size=sizeof(float);
      break;
    case '1':
      type->Size=sizeof(double);
      break;
    }
    s++; // varsize;
    break;
  case 'b':
  case 'i':
    switch(*s) {
    case '0':
      type->Type=CHARSYM;
      type->Size=sizeof(char);
      break;
    case '1':
      type->Type=SHORTSYM;
      type->Size=sizeof(short);
      break;
    case '2':
      type->Type=INTSYM;
      type->Size=sizeof(int);
      break;
    case '3':
      type->Type=INTSYM;
      type->Size=sizeof(long);
      break;
    case '4':
    case '5':
      type->Type=INTSYM;
      type->Size=2*sizeof(int);
      break;
    }
    s++; // varsize
    if (*s=='?') type->Signed=2;
    else type->Signed=(*s)-'0';
    s++; // signed
    break;
  case 'V':
    type->Type=VASYM;
    break;
  case 'v':
    type->Type=VOIDSYM;
    break;
  case 'S':
  case 'U':
    s--;
    type->Type=STRUCTSYM;
    type->Structtype=(struct JSX_struct *)*tb;
    JSX_struct_new(cx, obj, &s,tb,tt);
    break;
  case 'F':
    s--;
    type->Type=RETURNSYM;
    type->Func=(struct JSX_functype *)*tb;
    JSX_functype_new(cx, obj, &s, tb, tt);
    break;
  }

  *_s=s;
}

void JSX_functype_new(JSContext *cx, JSObject *obj, char **_s, char **tb, char **tt) {
  char *s=*_s;
  int i;
  struct JSX_functype *Func;

  s++; // 'F'

  Func=(struct JSX_functype *)*tb;
  memset(Func, 0, sizeof(struct JSX_functype));
  Func->Callconv=(*(s++)=='c')?CDECLSYM:STDCALLSYM;
  (*tb)+=sizeof(struct JSX_functype);
  if (*s!='v') {
    Func->Ret=(struct JSX_type *)*tb;
    JSX_type_new(cx, obj, &s,tb,tt); // return type
  } else {
    Func->Ret=0;
    s++;
  }
  Func->nArgs=atoi(s);
  s=strchr(s,' ')+1;
  Func->Args=(struct JSX_type **)*tb;
  (*tb)+=sizeof(struct JSX_type *)*Func->nArgs;
  
  for (i=0; i<Func->nArgs; i++) {
    Func->Args[i]=(struct JSX_type *)*tb;
    Func->Args[i]->Const=(*s)-'0';
    s++; // const
    JSX_type_new(cx, obj, &s, tb, tt); // param
  }
  Func->Elipsis=(*s)-'0';
  s++; // elipsis
  *_s=s;
}


size_t JSX_sizeof(JSContext * cx, JSObject * this, struct JSX_type *C, jsval *v) {
  size_t ret;

  if (C==NULL) { // unknown C size, depends on JS type
      if (JSVAL_IS_INT(*v) || JSVAL_IS_BOOLEAN(*v)) return sizeof(int32);
      if (JSVAL_IS_NUMBER(*v)) return sizeof(double);
      if (JSVAL_IS_STRING(*v)) return sizeof(char *);
      if (!JSVAL_IS_NULL(*v) && JSVAL_IS_OBJECT(*v)) return sizeof(void *);
      return sizeof(void *);
  }
  if (C->Dim==0) {
    if (C->Size) return C->Size;
    if (C->Type==VASYM) return sizeof( void * );
    if (C->Type==STRUCTSYM || C->Type==UNIONSYM) {
      struct JSX_struct *strct=C->Structtype;
      if (!strct) strct=JSX_lookupstruct(cx, this, C->Struct);
      if (!strct) return 0;
      else return strct->Sizeof;
    }
	if (C->Type==RETURNSYM) return sizeof(void *);
  }

  if (C->Array_size[0]==0) return sizeof(void *); // Is a pointer

  // Is an array

  // Dereference
  C->Array_size++;
  C->Dim--;
  ret=C->Array_size[-1]*JSX_sizeof(cx, this, C,v);
  // Rereference
  C->Dim++;
  C->Array_size--;
  return ret;
}

size_t JSX_align(JSContext * cx, JSObject * this, struct JSX_type *C, jsval *v) {
  size_t ret;

  if (C==NULL) { // unknown C size, depends on JS type
      if (JSVAL_IS_INT(*v) || JSVAL_IS_BOOLEAN(*v)) return sizeof(int32);
      if (JSVAL_IS_NUMBER(*v)) return sizeof(int32);
      if (JSVAL_IS_STRING(*v)) return sizeof(char *);
      if (!JSVAL_IS_NULL(*v) && JSVAL_IS_OBJECT(*v)) return sizeof(void *);
      return sizeof(void *);
  }
  if (C->Dim==0) {
    if (C->Size && C->Size<sizeof(int32)) return C->Size;
    if (C->Size) return sizeof(int32);
    if (C->Type==VASYM) return sizeof( void * );
    if (C->Type==STRUCTSYM || C->Type==UNIONSYM) {
      size_t ret=0;
      int i;

      struct JSX_struct *strct=C->Structtype;
      if (!strct) strct=JSX_lookupstruct(cx, this, C->Struct);
      if (!strct) return sizeof(int32); // guesswork
      for (i=0; i<strct->nMembers; i++) {
	size_t thisalign=JSX_align(cx, this, strct->Members[i], v);
	if (thisalign>ret) ret=thisalign;
      }
      return ret;
    }
  }

  if (C->Array_size[0]==0) return sizeof(void *); // Is a pointer

  // Is an array

  // Dereference
  C->Array_size++;
  C->Dim--;
  ret=JSX_align(cx, this, C,v);
  // Rereference
  C->Dim++;
  C->Array_size--;
  return ret;
}

int JSX_struct_member(struct JSX_struct *S, char *name) {
  int i;
  for (i=0; i<S->nMembers; i++) {
    if (strcmp(name, S->Members[i]->Name)==0) return i;
  }
  return -1;
}

struct JSX_struct *JSX_lookupstruct(JSContext *cx, JSObject *this, char *Structname) {
  jsval v;
  JSObject *obj;
  JSBool res;
  JSObject *cont;

  if (!this) return 0;
  res=JS_GetProperty(cx, this, "struct", &v);
  if (!res || v==JSVAL_VOID || !JSVAL_IS_OBJECT(v)) return 0;
  cont=JSVAL_TO_OBJECT(v);
  if (!cont) return 0;
  res=JS_GetProperty(cx, cont, Structname, &v);
  if (!res) return 0;

  if (JSVAL_IS_OBJECT(v) && !JSVAL_IS_NULL(v) &&
      (obj=JSVAL_TO_OBJECT(v)) && 
      JS_InstanceOf (cx, obj, &JSX_type_class, 0)) {
    struct JSX_type *type=JS_GetPrivate(cx, obj);
    return type->Structtype;
  }
  return 0;
}


void JS_DLL_CALLBACK JSX_Type_finalize(JSContext *cx, JSObject *obj);

JSClass JSX_type_class={
  "Type",
  JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,
  JS_PropertyStub,
  JS_PropertyStub,
  JS_PropertyStub,
  JS_EnumerateStub,
  JS_ResolveStub,
  JS_ConvertStub,
  JSX_Type_finalize
};

JSObject *JSX_type_proto;

/*
  Create a new type object from a string. Used in C wrappers
 */

#ifdef _WIN32
__declspec(dllexport)
#endif
JSObject *JSX_Type_new_x(JSContext *cx, JSObject *this, JSObject *object, char *s) {
  char *s2;
  int size;
  char *t;
  struct JSX_type *priv;
  char *tb, *tt;

  if (!object) object=JS_NewObject(cx, &JSX_type_class, JSX_type_proto, 0);

  // 1. find size
  s2=s;
  size=JSX_type_size_s(&s2);

  // 2. allocate memory
  priv=(struct JSX_type *)JS_malloc(cx,size);
  if (!priv) return JS_FALSE;
  t=(char *)priv;
  memset(t,0,size);
  
  // 3. fill data
  s2=s;
  tb=t;
  tt=t+size;
  JSX_type_new(cx, this, &s2, &tb, &tt);

  JS_SetPrivate(cx, object, priv);

  return object;
}

/*
  'new' method for the struct prototype
 */

JSBool JS_DLL_CALLBACK JSX_Type_new(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
  JSObject *object;
  JSObject *this=0;
  char *s;

  if (JS_FALSE==JS_InstanceOf( cx, obj, &JSX_type_class, 0)) { // not called with new
    object=JS_NewObject( cx, &JSX_type_class, 0, 0);
  } else {
    object=obj;
  }
    
  if (argc<1) return JS_FALSE;
  if (!JSVAL_IS_STRING(argv[0])) return JS_FALSE;
  s=JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));

  if (argc>=2 && JSVAL_IS_OBJECT(argv[1])) this=JSVAL_TO_OBJECT(argv[1]);
  // Second argument should be the _ object, containing a struct object.

  JSX_Type_new_x(cx, this, object, s);
    
  if (obj!=object) {
    *rval=OBJECT_TO_JSVAL(object);
  }
  return JS_TRUE;
}

void JS_DLL_CALLBACK JSX_Type_finalize(JSContext *cx, JSObject *obj) {
  struct JSX_type *p=JS_GetPrivate(cx, obj);
  JS_free(cx, p);
}
